from typing import Any, Dict

from flask import Flask
from flask_marshmallow import Marshmallow
from marshmallow import fields, post_load

from .model import User

__all__ = ["org_schema", "user_schema", "workspace_schema", "new_user_schema",
           "orgs_schema", "ma", "link_org_schema", "link_workspace_schema",
           "new_workspace_schema", "workspaces_schema", "orgs_schema_tiny",
           "org_schema_short", "workspaces_schema_tiny", "my_orgs_schema",
           "workspace_schema_short", "my_workspaces_schema",
           "experiments_schema", "my_experiments_schema", "setup_schemas",
           "my_executions_schema", "my_schedules_schema"]

ma = Marshmallow()


def setup_schemas(app: Flask):
    return ma.init_app(app)


class WorkspaceSchema(ma.Schema):
    class Meta:
        ordered = True
    id = fields.UUID(required=True)
    org_id = fields.UUID(required=True)
    name = fields.String(required=True)
    kind = fields.String(
        data_key="type",
        required=True,
        validate=lambda k: k in ("personal", "protected", "public")
    )
    owner = fields.Boolean(default=False)
    visibility = fields.Dict(
        keys=fields.Str(validate=lambda k: k in ("execution", "experiment")),
        values=fields.Dict(
            keys=fields.String(
                validate=lambda k: k in ("anonymous", "members")
            ),
            values=fields.String(
                validate=lambda v: v in (
                    "private", "protected", "public",
                    "none", "status", "full"
                )
            )
        )
    )
    url = ma.AbsoluteURLFor('workspace.get', workspace_id='<id>')
    links = ma.Hyperlinks({
        'self': ma.URLFor('workspace.get', workspace_id='<id>')
    })


class NewWorkspaceSchema(ma.Schema):
    name = fields.String(required=True)
    org = fields.UUID(required=True)
    kind = fields.String(
        data_key="type",
        missing="collaborative",
        validate=lambda k: k in ("personal", "collaborative")
    )
    visibility = fields.Dict(
        keys=fields.Str(validate=lambda k: k in ("execution", "experiment")),
        values=fields.Dict(
            keys=fields.String(
                validate=lambda k: k in ("anonymous", "members")
            ),
            values=fields.String(
                validate=lambda v: v in (
                    "private", "protected", "public",
                    "none", "status", "full"
                )
            )
        )
    )


class LinkWorkpaceSchema(ma.Schema):
    owner = fields.Boolean(default=False)


class OrganizationSchema(ma.Schema):
    class Meta:
        ordered = True
    id = fields.UUID(required=True)
    name = fields.String(required=True)
    owner = fields.Boolean(default=False)
    kind = fields.String(data_key="type", required=True)
    workspaces = fields.Nested(WorkspaceSchema, many=True)
    url = ma.AbsoluteURLFor('org.get', org_id='<id>')
    links = ma.Hyperlinks({
        'self': ma.URLFor('org.get', org_id='<id>')
    })


class NewOrgSchema(ma.Schema):
    name = fields.String(required=True)


class LinkOrgSchema(ma.Schema):
    owner = fields.Boolean(default=False)


class UserSchema(ma.Schema):
    class Meta:
        ordered = True
    id = fields.UUID(required=False)
    username = fields.String(required=True)
    orgs = fields.Nested(OrganizationSchema, many=True, dump_only=True)
    workspaces = fields.Nested(WorkspaceSchema, many=True, dump_only=True)
    url = ma.AbsoluteURLFor('user.get', user_id='<id>')
    links = ma.Hyperlinks({
        'self': ma.URLFor('user.get', user_id='<id>'),
    })

    @post_load
    def make_user(self, data: Dict[str, Any]):
        return User(**data)


class NewUserSchema(ma.Schema):
    username = fields.String(required=True)


class ExperimentSchema(ma.Schema):
    class Meta:
        ordered = True
    id = fields.UUID(required=True)
    user_id = fields.UUID(required=True)
    org_id = fields.UUID(required=True)
    workspace_id = fields.UUID(required=True)
    created_date = fields.DateTime()
    updated_date = fields.DateTime()


class ExecutionSchema(ma.Schema):
    class Meta:
        ordered = True
    id = fields.UUID(required=True)
    user_id = fields.UUID(required=True)
    org_id = fields.UUID(required=True)
    workspace_id = fields.UUID(required=True)
    execution_id = fields.UUID(required=True)


class ScheduleSchema(ma.Schema):
    class Meta:
        ordered = True
    id = fields.UUID(required=True)
    user_id = fields.UUID(required=True)
    org_id = fields.UUID(required=True)
    workspace_id = fields.UUID(required=True)
    experiment_id = fields.UUID(required=True)
    token_id = fields.UUID(required=True)
    scheduled = fields.DateTime(required=True)
    status = fields.String(required=True, default="pending")
    job_id = fields.UUID(required=False)


new_user_schema = NewUserSchema()
user_schema = UserSchema()
link_org_schema = LinkOrgSchema()
link_workspace_schema = LinkWorkpaceSchema()

org_schema = OrganizationSchema()
orgs_schema = OrganizationSchema(many=True)
orgs_schema_tiny = OrganizationSchema(
    many=True, exclude=('owner', 'workspaces'))
my_orgs_schema = OrganizationSchema(many=True, exclude=('workspaces', ))
org_schema_short = OrganizationSchema(exclude=('owner',))
new_org_schema = NewOrgSchema()

workspace_schema = WorkspaceSchema()
workspaces_schema = WorkspaceSchema(many=True)
workspaces_schema_tiny = WorkspaceSchema(
    many=True, exclude=('owner',))
workspace_schema_short = WorkspaceSchema(exclude=('owner',))
my_workspaces_schema = WorkspaceSchema(many=True)
new_workspace_schema = NewWorkspaceSchema()

experiments_schema = ExperimentSchema(many=True)
my_experiments_schema = ExperimentSchema(many=True)

my_executions_schema = ExecutionSchema(many=True)

my_schedules_schema = ScheduleSchema(many=True)



# -*- coding: utf-8 -*-
from typing import NoReturn, Union
import uuid
from uuid import UUID

from chaosplt_relational_storage.db import Base, get_secret_key
from flask_login import UserMixin
from sqlalchemy import Boolean, Column, DateTime, ForeignKey, \
    String, func
from sqlalchemy.orm import backref, relationship
from sqlalchemy.orm.session import Session
from sqlalchemy_utils import EncryptedType, UUIDType
from sqlalchemy_utils.types.encrypted.encrypted_type import AesEngine
from sqlalchemy_utils import JSONType as JSONB

__all__ = ["User", "UserInfo", "UserPrivacy"]


class User(Base, UserMixin):  # type: ignore
    __tablename__ = 'user'

    id = Column(
        UUIDType(binary=False), primary_key=True, default=uuid.uuid4)
    joined_dt = Column(DateTime(), server_default=func.now())
    closed_dt = Column(DateTime())
    inactive_dt = Column(DateTime())
    is_closed = Column(Boolean(name='is_closed'), default=False)
    is_active = Column(Boolean(name='is_active'), default=True)
    is_authenticated = Column(Boolean(name="is_authenticated"), default=False)
    is_anonymous = Column(Boolean(name="is_anonymous"), default=False)

    info = relationship(
        'UserInfo', backref='user', uselist=False,
        cascade="all, delete-orphan")
    privacy = relationship(
        'UserPrivacy', backref='user', uselist=False,
        cascade="all, delete-orphan")
    workspaces = relationship(
        'Workspace', secondary="workspaces_members",
        lazy='subquery', backref=backref(
            'users', lazy=True, passive_updates=False))
    orgs = relationship(
        'Org', secondary="orgs_members", lazy='subquery',
        backref=backref('users', lazy=True, passive_updates=False))
    # direct access to the unique personal org
    # but this org is also part of the many to many relationship
    personal_org = relationship(
        'Org', backref='user', uselist=False, cascade="all, delete-orphan")

    @staticmethod
    def load(user_id: Union[UUID, str], session: Session) -> 'User':
        return session.query(User).filter_by(id=user_id).first()

    @staticmethod
    def create(username: str, name: str, email: str,
               session: Session) -> 'User':
        info = UserInfo(
            username=username,
            fullname=name,
            details={
                "email": email
            })
        privacy = UserPrivacy()
        user = User(
            info=info,
            privacy=privacy,
            is_active=True,
            is_authenticated=True
        )

        session.add(info)
        session.add(privacy)
        session.add(user)
        return user

    @staticmethod
    def delete(user_id: Union[UUID, str], session: Session) -> NoReturn:
        user = session.query(User).filter_by(id=user_id).first()

        if user:
            session.delete(user)


class UserInfo(Base):  # type: ignore
    __tablename__ = 'user_info'

    id = Column(
        UUIDType(binary=False), primary_key=True, default=uuid.uuid4)
    user_id = Column(
        UUIDType(binary=False), ForeignKey('user.id'),
        nullable=False)
    last_updated = Column(
        DateTime(), server_default=func.now(),
        onupdate=func.current_timestamp())
    verified_email = Column(Boolean(name='verified_email'), default=False)

    # values for search purpose mostly
    username = Column(String, index=True, nullable=True)
    fullname = Column(String, index=True, nullable=True)

    details = Column(
        EncryptedType(
            JSONB, get_secret_key, AesEngine, 'pkcs5'))


class UserPrivacy(Base):  # type: ignore
    __tablename__ = 'user_privacy'

    id = Column(
        UUIDType(binary=False), primary_key=True, default=uuid.uuid4)
    user_id = Column(
        UUIDType(binary=False), ForeignKey('user.id'),
        nullable=False)
    last_changed = Column(DateTime(), server_default=func.now())
    details = Column(JSONB())



from typing import Any, Dict

from chaosplt_grpc import remote_channel
from chaosplt_grpc.experiment.client import create_registration, \
    get_by_id

from ..model import User

__all__ = ["ExperimentService"]


class ExperimentService:
    def __init__(self, config: Dict[str, Any]):
        self.exp_addr = config["grpc"]["experiment"]["addr"]

    def release(self):
        pass

    def create(self, username: str, name: str, email: str) -> User:
        with remote_channel(self.exp_addr) as channel:
            registration = create_registration(channel, username, name, email)
            return User(
                registration.id, registration.username, registration.name,
                registration.email, is_active=registration.is_active,
                is_authenticated=registration.is_authenticated,
                is_anonymous=registration.is_anonymous)

    def get(self, registration_id: str) -> User:
        with remote_channel(self.exp_addr) as channel:
            registration = get_by_id(channel, registration_id)
            return User(
                registration.id, registration.username, registration.name,
                registration.email, is_active=registration.is_active,
                is_authenticated=registration.is_authenticated,
                is_anonymous=registration.is_anonymous)



# -*- coding: utf-8 -*-
from enum import Enum

__all__ = ["Org", "OrgsMembers", "OrgType", "User", "UserInfo", "UserPrivacy",
           "Workspace", "WorkspacesMembers", "WorkspaceType",
           "ExperimentVisibility", "ExecutionVisibility"]


class ExperimentVisibility(Enum):
    private = "private"
    protected = "protected"
    public = "public"

    @staticmethod
    def is_valid_visibility(experiment_visibility: str):
        return experiment_visibility in (
            ExperimentVisibility.private.value,
            ExperimentVisibility.protected.value,
            ExperimentVisibility.public.value,
        )

    @staticmethod
    def get_visibility(experiment_visibility: str):
        if experiment_visibility == ExperimentVisibility.private.value:
            return ExperimentVisibility.personal

        if experiment_visibility == ExperimentVisibility.protected.value:
            return ExperimentVisibility.protected

        if experiment_visibility == ExperimentVisibility.public.value:
            return ExperimentVisibility.public


class ExecutionVisibility(Enum):
    none = "none"
    status = "status"
    full = "full"

    @staticmethod
    def is_valid_visibility(execution_visibility: str):
        return execution_visibility in (
            ExecutionVisibility.none.value,
            ExecutionVisibility.status.value,
            ExecutionVisibility.full.value,
        )

    @staticmethod
    def get_visibility(execution_visibility: str):
        if execution_visibility == ExecutionVisibility.none.value:
            return ExecutionVisibility.none

        if execution_visibility == ExecutionVisibility.status.value:
            return ExecutionVisibility.status

        if execution_visibility == ExecutionVisibility.full.value:
            return ExecutionVisibility.full


from .org import Org, OrgsMembers, OrgType  # noqa: E402
from .user import User, UserInfo, UserPrivacy  # noqa: E402
from .workspace import Workspace, WorkspacesMembers, WorkspaceType  # noqa: E402



# -*- coding: utf-8 -*-
from enum import Enum
from typing import List, NoReturn, Union
import uuid
from uuid import UUID

from chaosplt_relational_storage.db import Base
from sqlalchemy import Boolean, Column, DateTime, ForeignKey, \
    String, func
from sqlalchemy import Enum as EnumType
from sqlalchemy.orm import relationship
from sqlalchemy.orm.session import Session
from sqlalchemy_utils import UUIDType
from sqlalchemy_utils import JSONType as JSONB

from . import ExperimentVisibility, ExecutionVisibility
from .user import User

__all__ = ["OrgsMembers", "Org", "OrgType"]


class OrgsMembers(Base):  # type: ignore
    __tablename__ = "orgs_members"

    org_id = Column(
        UUIDType(binary=False), ForeignKey('org.id'), primary_key=True)
    user_id = Column(
        UUIDType(binary=False), ForeignKey('user.id'), primary_key=True)
    is_owner = Column(Boolean(name='is_owner'), default=False)
    user = relationship('User')
    organization = relationship('Org')

    @staticmethod
    def create(org: 'Org', user: 'User', owner: bool,
               session: Session) -> 'OrgsMembers':
        assoc = OrgsMembers(
            organization=org,
            user=user,
            is_owner=owner
        )
        session.add(assoc)
        return assoc

    @staticmethod
    def create_from_ids(org_id: Union[UUID, str], user_id: Union[UUID, str],
                        owner: bool, session: Session) -> 'OrgsMembers':
        assoc = OrgsMembers(
            org_id=org_id,
            user_id=user_id,
            is_owner=owner
        )
        session.add(assoc)
        return assoc

    @staticmethod
    def load(org_id: Union[UUID, str], user_id: Union[UUID, str],
             session: Session) -> 'OrgsMembers':
        return session.query(OrgsMembers).\
            filter_by(org_id=org_id).\
            filter_by(user_id=user_id).\
            first()


class OrgType(Enum):
    personal = "personal"
    collaborative = "collaborative"


DEFAULT_ORG_SETTINGS = {  # type: ignore
    "meta": {},
    "visibility": {
        "execution": {
            "anonymous": ExecutionVisibility.none.value,
            "members": ExecutionVisibility.full.value
        },
        "experiment": {
            "anonymous": ExperimentVisibility.public.value,
            "members": ExperimentVisibility.public.value,
        }
    }
}


class Org(Base):  # type: ignore
    __tablename__ = 'org'

    id = Column(
        UUIDType(binary=False), primary_key=True, default=uuid.uuid4)
    # only set when this is a personal org linked to a single account,
    # otherwise it's not set
    user_id = Column(
        UUIDType(binary=False), ForeignKey('user.id'), nullable=True)
    name = Column(String(), nullable=False, unique=True)
    name_lower = Column(String(), nullable=False, unique=True)
    kind = Column(
        EnumType(OrgType), nullable=False, default=OrgType.personal)
    created_on = Column(DateTime(), server_default=func.now())
    workspaces = relationship(
        'Workspace', backref='org', cascade="all, delete-orphan")
    settings = Column(
        JSONB(), nullable=False, default=DEFAULT_ORG_SETTINGS)

    @staticmethod
    def load_all(session: Session) -> List['Org']:
        return session.query(Org).all()

    @staticmethod
    def load(org_id: Union[UUID, str], session: Session) -> 'Org':
        return session.query(Org).filter_by(id=org_id).first()

    @staticmethod
    def load_by_name(org_name: str, session: Session) -> 'Org':
        name = org_name.lower()
        return session.query(Org).filter_by(name_lower=name).first()

    @staticmethod
    def create_personal(user: User, org_name: str,
                        session: Session) -> 'Org':
        org = Org(
            user=user,
            name=org_name,
            name_lower=org_name.lower()
        )

        session.add(org)
        return org

    @staticmethod
    def create(org_name: str, session: Session) -> 'Org':
        org = Org(
            name=org_name,
            name_lower=org_name.lower(),
            kind=OrgType.collaborative
        )

        session.add(org)
        return org

    @staticmethod
    def delete(org_id: Union[UUID, str], session: Session) -> NoReturn:
        org = session.query(Org).filter_by(id=org_id).first()

        if org:
            session.delete(org)

    def get_workspace_by_name(self, workspace_name: str,
                              session: Session) -> 'Workspace':  # noqa: E0602
        for w in self.workspaces:
            if w.name == workspace_name:
                return w

    def get_workspace_by_id(self, workspace_id: str,
                            session: Session) -> 'Workspace':  # noqa: E0602
        for w in self.workspaces:
            if w.id == workspace_id:
                return w



# -*- coding: utf-8 -*-
import logging
from logging import StreamHandler
import os
from typing import Any, Dict

from flask import Blueprint, Flask, after_this_request, request, Response
from flask_caching import Cache

from chaosplt_account.auth import setup_jwt, setup_login
from chaosplt_account.schemas import setup_schemas
from chaosplt_account.service import Services
from chaosplt_account.storage import AccountStorage

from .org import api as org_api
from .user import api as user_api
from .workspace import api as workspace_api

__all__ = ["create_api", "cleanup_api", "serve_api"]


def create_api(config: Dict[str, Any]) -> Flask:
    app = Flask(__name__)

    app.url_map.strict_slashes = False
    app.debug = True if os.getenv('CHAOSHUB_DEBUG') else False

    logger = logging.getLogger('flask.app')
    logger.propagate = False

    app.config["SECRET_KEY"] = os.getenv("SECRET_KEY")
    app.config["SQLALCHEMY_DATABASE_URI"] = os.getenv("DATABASE_URI")
    app.config["JWT_SECRET_KEY"] = os.getenv("JWT_SECRET_KEY")
    app.secret_key = app.config["SECRET_KEY"]

    app.config["GRPC_ADDR"] = "{}:{}".format(
        os.getenv("GRPC_LISTEN_ADDR"), os.getenv("GRPC_LISTEN_PORT"))

    app.config["CACHE_TYPE"] = os.getenv("CACHE_TYPE", "simple")
    if app.config["CACHE_TYPE"] == "redis":
        app.config["CACHE_REDIS_HOST"] = os.getenv("CACHE_REDIS_HOST")
        app.config["CACHE_REDIS_PORT"] = os.getenv("CACHE_REDIS_PORT", 6379)

    setup_jwt(app)
    setup_schemas(app)
    setup_login(app, from_jwt=True)

    return app


def cleanup_api(app: Flask):
    pass


def serve_api(app: Flask, cache: Cache, services: Services,
              storage: AccountStorage, config: Dict[str, Any],
              mount_point: str = '', log_handler: StreamHandler = None):
    register_api(app, cache, services, storage)


###############################################################################
# Internals
###############################################################################
def register_api(app: Flask, cache: Cache, services, storage: AccountStorage):
    patch_request(user_api, services, storage)
    patch_request(org_api, services, storage)
    patch_request(workspace_api, services, storage)

    app.register_blueprint(user_api, url_prefix="/users")
    app.register_blueprint(org_api, url_prefix="/organizations")
    app.register_blueprint(workspace_api, url_prefix="/workspaces")


def patch_request(bp: Blueprint, services, storage: AccountStorage):
    @bp.before_request
    def prepare_request():
        request.services = services
        request.storage = storage

        @after_this_request
        def clean_request(response: Response):
            request.services = None
            request.storage = None
            return response



from typing import Dict, List, NoReturn, Union
from uuid import UUID

from chaosplt_account.model import Organization, User, Workspace
from chaosplt_relational_storage import RelationalStorage
from chaosplt_relational_storage.db import orm_session

from .interface import BaseOrganizationService, BaseUserService, \
    BaseRegistrationService, BaseWorkspaceService
from .model import User as UserModel, \
    Org as OrgModel, Workspace as WorkspaceModel, \
    OrgsMembers as OrgsMembersAssociation, \
    WorkspacesMembers as WorkspaceMembersAssociation, WorkspaceType


__all__ = ["UserService", "OrgService", "WorkspaceService",
           "RegistrationService"]


class UserService(BaseUserService):
    def __init__(self, driver: RelationalStorage):
        self.driver = driver

    def get_bare(self, user_id: Union[UUID, str]) -> User:
        with orm_session() as session:
            user = UserModel.load(user_id, session=session)
            if not user:
                return

            return User(
                id=user.id,
                name=user.info.fullname,
                username=user.info.username,
                email=user.info.details.get('email'),
                is_authenticated=user.is_authenticated,
                is_active=user.is_active,
                is_anonymous=user.is_anonymous,
                orgs=[],
                workspaces=[]
            )

    def get(self, user_id: Union[UUID, str]) -> User:
        with orm_session() as session:
            user = UserModel.load(user_id, session=session)
            if not user:
                return

            user_orgs = user.orgs or []
            orgs = []
            for org in user_orgs:
                org_id = org.id
                assoc = OrgsMembersAssociation.load(
                    org_id, user_id, session=session)
                orgs.append(
                    Organization(
                        id=org_id,
                        name=org.name,
                        owner=assoc.is_owner,
                        kind=org.kind.value,
                        workspaces=[]
                    )
                )

            user_workspaces = user.workspaces or []
            workspaces = []
            for workspace in user_workspaces:
                workspace_id = workspace.id
                assoc = WorkspaceMembersAssociation.load(
                    workspace_id, user_id, session=session)
                settings = workspace.settings

                workspaces.append(
                    Workspace(
                        id=workspace_id,
                        org_id=workspace.org_id,
                        name=workspace.name,
                        kind=workspace.kind.value,
                        owner=assoc.is_owner,
                        visibility=settings.get("visibility")
                    )
                )

            if user:
                return User(
                    id=user.id,
                    name=user.info.fullname,
                    username=user.info.username,
                    email=user.info.details.get('email'),
                    is_authenticated=user.is_authenticated,
                    is_active=user.is_active,
                    is_anonymous=user.is_anonymous,
                    orgs=orgs,
                    workspaces=workspaces
                )

    def create(self, username: str, name: str, email: str) -> User:
        with orm_session() as session:
            user = UserModel.create(username, name, email, session=session)
            org = OrgModel.create_personal(user, username, session=session)
            workspace = WorkspaceModel.create(
                user, org, username, WorkspaceType.personal.value, None,
                session=session)
            session.flush()

            settings = workspace.settings

            OrgsMembersAssociation.create(
                org=org, user=user, owner=True, session=session)
            WorkspaceMembersAssociation.create(
                workspace=workspace, user=user, owner=True, session=session)

            return User(
                id=user.id,
                name=user.info.fullname,
                username=user.info.username,
                email=user.info.details.get('email'),
                is_authenticated=user.is_authenticated,
                is_active=user.is_active,
                is_anonymous=user.is_anonymous,
                orgs=[
                    Organization(
                        id=org.id,
                        name=org.name,
                        kind=org.kind.value,
                        owner=True
                    )
                ],
                workspaces=[
                    Workspace(
                        id=workspace.id,
                        org_id=workspace.org_id,
                        name=workspace.name,
                        kind=workspace.kind.value,
                        owner=True,
                        visibility=settings.get("visibility")
                    )
                ]
            )

    def delete(self, user_id: Union[UUID, str]) -> NoReturn:
        with orm_session() as session:
            UserModel.delete(user_id, session=session)

    def add_org(self, user_id: Union[UUID, str],
                org_id: Union[UUID, str], owner: bool = False) -> NoReturn:
        with orm_session() as session:
            OrgsMembersAssociation.create_from_ids(
                org_id=org_id, user_id=user_id, owner=owner, session=session)

    def remove_org(self, user_id: Union[UUID, str],
                   org_id: Union[UUID, str]) -> NoReturn:
        with orm_session() as session:
            user = UserModel.load(user_id, session=session)
            org = OrgModel.load(org_id, session=session)
            if not user.orgs:
                return

            for o in user.orgs:
                if o.id == org.id:
                    user.orgs.remove(o)
                    break

    def add_workspace(self, user_id: Union[UUID, str],
                      workspace_id: Union[UUID, str],
                      owner: bool = False) -> NoReturn:
        with orm_session() as session:
            WorkspaceMembersAssociation.create_from_ids(
                workspace_id=workspace_id, user_id=user_id, owner=owner,
                session=session)

    def remove_workspace(self, user_id: Union[UUID, str],
                         workspace_id: Union[UUID, str]) -> NoReturn:
        with orm_session() as session:
            user = UserModel.load(user_id, session=session)
            workspace = WorkspaceModel.load(workspace_id, session=session)
            if not user.workspaces:
                return

            for w in user.workspaces:
                if w.id == workspace.id:
                    user.workspaces.remove(w)
                    break


class OrgService(BaseOrganizationService):
    def __init__(self, driver: RelationalStorage):
        self.driver = driver

    def list_all(self) -> List[Organization]:
        with orm_session() as session:
            orgs = []
            for org in OrgModel.load_all(session):
                orgs.append(
                    Organization(
                        id=org.id,
                        name=org.name,
                        kind=org.kind.value
                    )
                )
            return orgs

    def get(self, org_id: Union[UUID, str]) -> Organization:
        with orm_session() as session:
            org = OrgModel.load(org_id, session=session)
            if org:
                return Organization(
                    id=org.id,
                    name=org.name,
                    kind=org.kind.value,
                    workspaces=[
                        Workspace(
                            id=workspace.id,
                            org_id=workspace.org_id,
                            name=workspace.name,
                            kind=workspace.kind.value,
                            visibility=workspace.settings.get('visibility')
                        ) for workspace in org.workspaces
                    ]
                )

    def create(self, name: str, user_id: Union[UUID, str]) -> Organization:
        with orm_session() as session:
            user = UserModel.load(user_id, session=session)
            org = OrgModel.create(name, session=session)

            OrgsMembersAssociation.create(
                org=org, user=user, owner=True, session=session)
            session.flush()

            return Organization(
                id=org.id,
                name=org.name,
                kind=org.kind.value,
                workspaces=[]
            )

    def delete(self, org_id: Union[UUID, str]) -> NoReturn:
        with orm_session() as session:
            OrgModel.delete(org_id, session=session)

    def add_workspace(self, org_id: Union[UUID, str],
                      workspace_id: Union[UUID, str],
                      owner: bool = False) -> NoReturn:
        with orm_session() as session:
            WorkspaceMembersAssociation.create_from_ids(
                org_id=org_id, workspace_id=workspace_id, owner=owner,
                session=session)

    def remove_workspace(self, org_id: Union[UUID, str],
                         workspace_id: Union[UUID, str]) -> NoReturn:
        with orm_session() as session:
            org = OrgModel.load(org_id, session=session)
            workspace = WorkspaceModel.load(org_id, session=session)

            for w in org.workspaces:
                if w.id == workspace.id:
                    org.workspaces.remove(workspace)
                    break

    def has_org_by_name(self, org_name: str) -> bool:
        with orm_session() as session:
            org = OrgModel.load_by_name(org_name, session=session)
            return org is not None

    def has_org_by_id(self, org_id: Union[UUID, str]) -> bool:
        with orm_session() as session:
            org = OrgModel.load(org_id, session=session)
            return org is not None

    def has_workspace_by_name(self, org_id: Union[UUID, str],
                              workspace_name: str) -> bool:
        with orm_session() as session:
            org = OrgModel.load(org_id, session=session)
            workspace = org.get_workspace_by_name(
                workspace_name, session=session)
            return workspace is not None

    def has_workspace_by_id(self, org_id: Union[UUID, str],
                            workspace_id: Union[UUID, str]) -> bool:
        with orm_session() as session:
            org = OrgModel.load(org_id, session=session)
            workspace = org.get_workspace_by_id(
                workspace_id, session=session)
            return workspace is not None


class WorkspaceService(BaseWorkspaceService):
    def __init__(self, driver: RelationalStorage):
        self.driver = driver

    def list_all(self) -> List[Workspace]:
        with orm_session() as session:
            workspaces = []
            for workspace in WorkspaceModel.load_all(session):
                settings = workspace.settings

                workspaces.append(
                    Workspace(
                        id=workspace.id,
                        org_id=workspace.org_id,
                        name=workspace.name,
                        kind=workspace.kind.value,
                        visibility=settings.get("visibility", {})
                    )
                )
            return workspaces

    def get(self, workspace_id: Union[UUID, str]) -> Workspace:
        with orm_session() as session:
            workspace = WorkspaceModel.load(workspace_id, session=session)
            if workspace:
                return Workspace(
                    id=workspace.id,
                    name=workspace.name,
                    org_id=workspace.org_id,
                    kind=workspace.kind.value,
                    visibility=workspace.settings.get('visibility')
                )

    def create(self, name: str, org_id: Union[UUID, str],
               user_id: Union[UUID, str], workspace_type: str,
               visibility: Dict[str, str]) -> Workspace:
        with orm_session() as session:
            user = UserModel.load(user_id, session=session)
            org = OrgModel.load(org_id, session=session)
            workspace = WorkspaceModel.create(
                user, org, name, workspace_type, visibility, session=session)
            session.flush()

            settings = workspace.settings

            WorkspaceMembersAssociation.create(
                workspace=workspace, user=user, owner=True, session=session)

            return Workspace(
                id=workspace.id,
                name=workspace.name,
                kind=workspace.kind.value,
                org_id=org.id,
                owner=True,
                visibility=settings.get("visibility", {})
            )

    def delete(self, workspace_id: Union[UUID, str]) -> NoReturn:
        with orm_session() as session:
            WorkspaceModel.delete(workspace_id, session=session)


class RegistrationService(UserService, BaseRegistrationService):
    def __init__(self, driver: RelationalStorage):
        UserService.__init__(self, driver)



from abc import ABC, abstractmethod
from uuid import UUID
from typing import Dict, List, NoReturn, Union

import attr
from chaosplt_account.model import User, Organization, Workspace

__all__ = ["BaseAccountStorage"]


class BaseRegistrationService(ABC):
    @abstractmethod
    def get(self, user_id: Union[UUID, str]) -> User:
        raise NotImplementedError()

    @abstractmethod
    def create(self, username: str, name: str, email: str) -> User:
        raise NotImplementedError()

    @abstractmethod
    def delete(self, user_id: Union[UUID, str]) -> NoReturn:
        raise NotImplementedError()


class BaseUserService(ABC):
    @abstractmethod
    def get(self, user_id: Union[UUID, str]) -> User:
        raise NotImplementedError()

    @abstractmethod
    def get_bare(self, user_id: Union[UUID, str]) -> User:
        raise NotImplementedError()

    @abstractmethod
    def create(self, username: str, name: str, email: str) -> User:
        raise NotImplementedError()

    @abstractmethod
    def delete(self, user_id: Union[UUID, str]) -> NoReturn:
        raise NotImplementedError()

    @abstractmethod
    def add_org(self, user_id: Union[UUID, str],
                org_id: Union[UUID, str], owner: bool = False) -> NoReturn:
        raise NotImplementedError()

    @abstractmethod
    def remove_org(self, user_id: Union[UUID, str],
                   org_id: Union[UUID, str]) -> NoReturn:
        raise NotImplementedError()

    @abstractmethod
    def add_workspace(self, user_id: Union[UUID, str],
                      workspace_id: Union[UUID, str],
                      owner: bool = False) -> NoReturn:
        raise NotImplementedError()

    @abstractmethod
    def remove_workspace(self, user_id: Union[UUID, str],
                         workspace_id: Union[UUID, str]) -> NoReturn:
        raise NotImplementedError()


class BaseOrganizationService(ABC):
    @abstractmethod
    def list_all(self) -> List[Organization]:
        raise NotImplementedError()

    @abstractmethod
    def get(self, org_id: Union[UUID, str]) -> Organization:
        raise NotImplementedError()

    @abstractmethod
    def create(self, name: str, user_id: Union[UUID, str]) -> Organization:
        raise NotImplementedError()

    @abstractmethod
    def delete(self, org_id: Union[UUID, str]) -> NoReturn:
        raise NotImplementedError()

    @abstractmethod
    def add_workspace(self, org_id: Union[UUID, str],
                      workspace_id: Union[UUID, str]) -> NoReturn:
        raise NotImplementedError()

    @abstractmethod
    def remove_workspace(self, user_id: Union[UUID, str],
                         org_id: Union[UUID, str]) -> NoReturn:
        raise NotImplementedError()

    @abstractmethod
    def has_org_by_name(self, org_name: str = None) -> bool:
        raise NotImplementedError()

    @abstractmethod
    def has_org_by_id(self, org_id: Union[UUID, str] = None) -> bool:
        raise NotImplementedError()

    @abstractmethod
    def has_workspace_by_name(self, org_id: Union[UUID, str],
                              workspace_name: str = None) -> bool:
        raise NotImplementedError()

    @abstractmethod
    def has_workspace_by_id(self, org_id: Union[UUID, str],
                            workspace_id: Union[UUID, str] = None) -> bool:
        raise NotImplementedError()


class BaseWorkspaceService(ABC):
    @abstractmethod
    def list_all(self) -> List[Workspace]:
        raise NotImplementedError()

    @abstractmethod
    def get(self, workspace_id: Union[UUID, str]) -> Workspace:
        raise NotImplementedError()

    @abstractmethod
    def create(self, name: str, org_id: Union[UUID, str],
               user_id: Union[UUID, str], workspace_type: str,
               visibility: Dict[str, Dict[str, str]]) -> Workspace:
        raise NotImplementedError()

    @abstractmethod
    def delete(self, workspace_id: Union[UUID, str]) -> NoReturn:
        raise NotImplementedError()


@attr.s
class BaseAccountStorage:
    user: BaseUserService = attr.ib()
    org: BaseOrganizationService = attr.ib()
    workspace: BaseWorkspaceService = attr.ib()
    registration: BaseRegistrationService = attr.ib()



# -*- coding: utf-8 -*-
from uuid import UUID

from flask import abort, Blueprint, jsonify, request
from flask_login import current_user, login_required
from marshmallow import ValidationError

from chaosplt_account.schemas import new_user_schema, user_schema, \
    link_org_schema, link_workspace_schema, my_orgs_schema, \
    my_workspaces_schema, my_experiments_schema, my_executions_schema, \
    my_schedules_schema

__all__ = ["api"]

api = Blueprint("user", __name__)


@api.route('', methods=['POST'])
@login_required
def new():
    try:
        payload = new_user_schema.load(request.json)
    except ValidationError as err:
        return jsonify(err.messages), 422
    user = request.storage.user.create(payload["username"])
    return user_schema.jsonify(user)


@api.route('<uuid:user_id>', methods=['GET'])
@login_required
def get(user_id: UUID):
    user = request.storage.user.get(user_id)
    if not user:
        return abort(404)
    return user_schema.jsonify(user)


@api.route('<uuid:user_id>', methods=['DELETE'])
@login_required
def delete(user_id: UUID):
    request.storage.user.delete(user_id)
    return "", 204


@api.route('<uuid:user_id>/organizations', methods=['GET'])
@login_required
def get_user_orgs(user_id: UUID):
    user = request.storage.user.get(user_id)
    if not user:
        return abort(404)
    return my_orgs_schema.jsonify(user.orgs)


@api.route('<uuid:user_id>/workspaces', methods=['GET'])
@login_required
def get_user_workspaces(user_id: UUID):
    user = request.storage.user.get(user_id)
    if not user:
        return abort(404)
    return my_workspaces_schema.jsonify(user.workspaces)


@api.route('<uuid:user_id>/organizations/<uuid:org_id>', methods=['PUT'])
@login_required
def link_org_to_user(user_id: UUID, org_id: UUID):
    try:
        payload = link_org_schema.load(request.json)
    except ValidationError as err:
        return jsonify(err.messages), 422

    user = request.storage.user.get(user_id)
    if not user:
        return jsonify({
            "message": "Unknown user"
        }), 422

    org = request.storage.org.get(org_id)
    if not org:
        return jsonify({
            "message": "Unknown organization"
        }), 422

    current_user

    request.storage.user.add_org(
        user_id, org_id, owner=payload["owner"])
    return "", 204


@api.route('<uuid:user_id>/organizations/<uuid:org_id>', methods=['DELETE'])
@login_required
def unlink_org_from_user(user_id: UUID, org_id: UUID):
    user = request.services.account.user.get(user_id)
    if not user:
        return jsonify({
            "message": "Unknown user"
        }), 422

    org = request.storage.org.get(org_id)
    if not org:
        return jsonify({
            "message": "Unknown organization"
        }), 422

    request.storage.user.remove_org(user_id, org_id)
    return "", 204


@api.route('<uuid:user_id>/workspaces/<uuid:workspace_id>', methods=['PUT'])
@login_required
def link_workspace_to_user(user_id: UUID, workspace_id: UUID):
    try:
        payload = link_workspace_schema.load(request.json)
    except ValidationError as err:
        return jsonify(err.messages), 422

    user = request.storage.user.get(user_id)
    if not user:
        return jsonify({
            "message": "Unknown user"
        }), 422

    workspace = request.storage.workspace.get(workspace_id)
    if not workspace:
        return jsonify({
            "message": "Unknown workspace"
        }), 422

    request.storage.user.add_workspace(
        user_id, workspace_id, owner=payload["owner"])
    return "", 204


@api.route('<uuid:user_id>/workspaces/<uuid:workspace_id>', methods=['DELETE'])
@login_required
def unlink_workspace_from_user(user_id: UUID, workspace_id: UUID):
    user = request.storage.user.get(user_id)
    if not user:
        return jsonify({
            "message": "Unknown user"
        }), 422

    workspace = request.storage.workspace.get(workspace_id)
    if not workspace:
        return jsonify({
            "message": "Unknown workspace"
        }), 422

    request.storage.user.remove_workspace(user_id, workspace_id)
    return "", 204


@api.route('<uuid:user_id>/experiments', methods=['GET'])
@login_required
def get_user_experiments(user_id: UUID):
    experiments = request.services.experiment.list_by_user(user_id)
    return my_experiments_schema.jsonify(experiments)


@api.route('<uuid:user_id>/executions', methods=['GET'])
@login_required
def get_user_executions(user_id: UUID):
    executions = request.services.execution.get_by_user(user_id)
    return my_executions_schema.jsonify(executions)


@api.route('<uuid:user_id>/schedules', methods=['GET'])
@login_required
def get_user_schedules(user_id: UUID):
    schedules = request.services.scheduling.get_by_user(user_id)
    return my_schedules_schema.jsonify(schedules)



# -*- coding: utf-8 -*-
from uuid import UUID

from flask import abort, Blueprint, jsonify, request
from flask_login import current_user, login_required
from marshmallow import ValidationError

from chaosplt_account.schemas import new_workspace_schema, workspace_schema, \
    workspaces_schema_tiny, workspace_schema_short, \
    experiments_schema

__all__ = ["api"]

api = Blueprint("workspace", __name__)


@api.route('', methods=['GET'])
@login_required
def list_all():
    workspaces = request.services.account.workspace.list_all()
    return workspaces_schema_tiny.jsonify(workspaces)


@api.route('', methods=['POST'])
@login_required
def new():
    user_id = current_user.id
    try:
        payload = new_workspace_schema.load(request.json)
    except ValidationError as err:
        return jsonify(err.messages), 422

    workspace_name = payload["name"]
    workspace_type = payload["kind"]
    workspace_visibility = payload["visibility"]
    org_id = payload["org"]
    workspace_svc = request.services.account.workspace

    has_org = request.services.account.org.has_org(org_id=org_id)
    if not has_org:
        return jsonify({
            "message": "Unknown organization"
        }), 422

    has_workspace = request.services.account.org.has_workspace(
        org_id, workspace_name)
    if has_workspace:
        return jsonify({
            "message": "Name already used in this organization"
        }), 409

    workspace = workspace_svc.create(
        workspace_name, org_id, user_id, workspace_type, workspace_visibility)
    return workspace_schema.jsonify(workspace)


@api.route('<uuid:workspace_id>', methods=['GET'])
@login_required
def get(workspace_id: UUID):
    workspace = request.services.account.workspace.get(workspace_id)
    if not workspace:
        return abort(404)
    return workspace_schema_short.jsonify(workspace)


@api.route('<uuid:workspace_id>', methods=['DELETE'])
@login_required
def delete(workspace_id: UUID):
    request.services.account.workspace.delete(workspace_id)
    return "", 204


@api.route('<uuid:workspace_id>/experiments', methods=['GET'])
@login_required
def get_experiments(workspace_id: UUID):
    workspace = request.services.account.workspace.get(workspace_id)
    if not workspace:
        return abort(404)

    experiments = request.services.experiment.get_by_workspace(workspace_id)
    return experiments_schema.jsonify(experiments)



import logging
from typing import NoReturn

from chaosplt_grpc.registration.message import Registration
from chaosplt_grpc.registration.server import \
    RegistrationService as GRPCRegistrationService

from ..storage import AccountStorage

__all__ = ["RegistrationRPC"]
logger = logging.getLogger("chaosplatform")


class RegistrationRPC(GRPCRegistrationService):
    def __init__(self, storage: AccountStorage):
        GRPCRegistrationService.__init__(self)
        self.storage = storage

    def create_registration(self, username: str, name: str,
                            email: str) -> Registration:
        user = self.storage.registration.create(username, name, email)
        logger.info("User {} is registered".format(user.id))
        return Registration(
            id=str(user.id),
            username=user.username,
            name=user.name,
            email=user.email,
            is_active=True,
            is_authenticated=True,
            is_anonymous=False
        )

    def delete_registration(self, registration_id: str) -> NoReturn:
        logger.info("User {} now deleted".format(registration_id))
        self.storage.registration.delete(registration_id)

    def get_by_id(self, registration_id: str) -> Registration:
        user = self.storage.registration.get(registration_id)
        if not user:
            logger.info("User {} does not exist".format(registration_id))
            return None

        return Registration(
            id=str(user.id),
            username=user.username,
            name=user.name,
            email=user.email,
            is_active=user.is_active,
            is_authenticated=user.is_authenticated,
            is_anonymous=user.is_active
        )



# -*- coding: utf-8 -*-
from enum import Enum
from typing import Dict, List, NoReturn, Union
import uuid
from uuid import UUID

from chaosplt_relational_storage.db import Base
from sqlalchemy import Boolean, Column, ForeignKey, String, UniqueConstraint
from sqlalchemy import Enum as EnumType
from sqlalchemy.orm import relationship
from sqlalchemy.orm.session import Session
from sqlalchemy_utils import UUIDType
from sqlalchemy_utils import JSONType as JSONB

from . import ExperimentVisibility, ExecutionVisibility
from .org import Org
from .user import User

__all__ = ["Workspace", "WorkspacesMembers", "WorkspaceType"]


class WorkspacesMembers(Base):  # type: ignore
    __tablename__ = "workspaces_members"

    workspace_id = Column(
        UUIDType(binary=False), ForeignKey('workspace.id'), primary_key=True)
    user_id = Column(
        UUIDType(binary=False), ForeignKey('user.id'), primary_key=True)
    is_owner = Column(Boolean(name='is_owner'), default=False)
    account = relationship('User')
    workspace = relationship('Workspace')

    @staticmethod
    def create(workspace: 'Workspace', user: 'User', owner: bool,
               session: Session) -> 'WorkspacesMembers':
        assoc = WorkspacesMembers(
            workspace=workspace,
            account=user,
            is_owner=owner
        )
        session.add(assoc)
        return assoc

    @staticmethod
    def create_from_ids(workspace_id: Union[UUID, str],
                        user_id: Union[UUID, str], owner: bool,
                        session: Session) -> 'WorkspacesMembers':
        assoc = WorkspacesMembers(
            workspace_id=workspace_id,
            user_id=user_id,
            is_owner=owner
        )
        session.add(assoc)
        return assoc

    @staticmethod
    def load(workspace_id: Union[UUID, str], user_id: Union[UUID, str],
             session: Session) -> 'WorkspacesMembers':
        return session.query(WorkspacesMembers).\
            filter_by(workspace_id=workspace_id).\
            filter_by(user_id=user_id).\
            first()


class WorkspaceType(Enum):
    personal = "personal"
    protected = "protected"
    public = "public"

    @staticmethod
    def is_valid_type(workspace_type: str):
        return workspace_type in (
            WorkspaceType.personal.value,
            WorkspaceType.protected.value,
            WorkspaceType.public.value,
        )

    @staticmethod
    def get_type(workspace_type: str):
        if workspace_type == WorkspaceType.personal.value:
            return WorkspaceType.personal

        if workspace_type == WorkspaceType.protected.value:
            return WorkspaceType.protected

        if workspace_type == WorkspaceType.public.value:
            return WorkspaceType.public


DEFAULT_WORKSPACE_SETTINGS = {
    "visibility": {
        "execution": {
            "anonymous": ExecutionVisibility.none.value,
            "members": ExecutionVisibility.full.value
        },
        "experiment": {
            "anonymous": ExperimentVisibility.public.value,
            "members": ExperimentVisibility.public.value,
        }
    }
}


class Workspace(Base):  # type: ignore
    __tablename__ = "workspace"
    __table_args__ = (
        UniqueConstraint(
            'name', 'org_id'
        ),
    )

    id = Column(
        UUIDType(binary=False), primary_key=True, default=uuid.uuid4)
    name = Column(String(), nullable=False)
    name_lower = Column(String(), nullable=False)
    kind = Column(
        EnumType(WorkspaceType), nullable=False,
        default=WorkspaceType.personal)
    org_id = Column(
        UUIDType(binary=False), ForeignKey('org.id'), nullable=False)
    settings = Column(
        JSONB(), nullable=False, default=DEFAULT_WORKSPACE_SETTINGS)

    @staticmethod
    def load_all(session: Session) -> List['Workspace']:
        return session.query(Workspace).all()

    @staticmethod
    def load(workspace_id: Union[UUID, str], session: Session) -> 'Org':
        return session.query(Workspace).filter_by(id=workspace_id).first()

    @staticmethod
    def create(user: User, org: Org, workspace_name: str, workspace_type: str,
               visibility: Dict[str, Dict[str, str]],
               session: Session) -> 'Workspace':

        visibility = visibility or \
            DEFAULT_WORKSPACE_SETTINGS["visibility"].copy()
        workspace = Workspace(
            org=org,
            name=workspace_name,
            name_lower=workspace_name.lower(),
            kind=WorkspaceType.get_type(workspace_type),
            settings={
                "visibility": visibility
            }
        )

        session.add(workspace)
        return workspace

    @staticmethod
    def delete(workspace_id: Union[UUID, str], session: Session) -> NoReturn:
        workspace = session.query(Workspace).filter_by(id=workspace_id).first()

        if workspace:
            session.delete(workspace)



# -*- coding: utf-8 -*-
import logging
from logging import StreamHandler
import os
from typing import Any, Dict

from flask import Flask
from flask_caching import Cache

from chaosplt_account.auth import setup_login
from chaosplt_account.service import Services
from chaosplt_account.storage import AccountStorage

__all__ = ["create_app", "cleanup_app", "serve_app"]


def create_app(config: Dict[str, Any]) -> Flask:
    app = Flask(__name__)

    app.url_map.strict_slashes = False
    app.debug = True if os.getenv('CHAOSHUB_DEBUG') else False

    logger = logging.getLogger('flask.app')
    logger.propagate = False

    app.config["SECRET_KEY"] = os.getenv("SECRET_KEY")
    app.config["SQLALCHEMY_DATABASE_URI"] = os.getenv("DATABASE_URI")
    app.config["JWT_SECRET_KEY"] = os.getenv("JWT_SECRET_KEY")
    app.secret_key = app.config["SECRET_KEY"]

    app.config["GRPC_ADDR"] = "{}:{}".format(
        os.getenv("GRPC_LISTEN_ADDR"), os.getenv("GRPC_LISTEN_PORT"))

    app.config["CACHE_TYPE"] = os.getenv("CACHE_TYPE", "simple")
    if app.config["CACHE_TYPE"] == "redis":
        app.config["CACHE_REDIS_HOST"] = os.getenv("CACHE_REDIS_HOST")
        app.config["CACHE_REDIS_PORT"] = os.getenv("CACHE_REDIS_PORT", 6379)

    setup_login(app, from_session=True)

    return app


def cleanup_app(app: Flask):
    pass


def serve_app(app: Flask, cache: Cache, services: Services,
              storage: AccountStorage, config: Dict[str, Any],
              mount_point: str = '/account',
              log_handler: StreamHandler = None):
    pass



from typing import Any, Dict, NoReturn

import attr

__all__ = ["initialize_services", "shutdown_services", "services", "Services"]


@attr.s
class Services:
    pass


def initialize_services(services: Services, config: Dict[str, Any]):
    pass


def shutdown_services(services: Services) -> NoReturn:
    pass



from typing import Dict, List, Union
from uuid import UUID

import attr

__all__ = ["Organization", "User", "Workspace"]


@attr.s
class Workspace:
    id: UUID = attr.ib()
    name: str = attr.ib()
    org_id: UUID = attr.ib()
    kind: str = attr.ib()
    visibility: Dict[str, str] = attr.ib()
    owner: bool = attr.ib(default=False)


@attr.s
class Organization:
    id: UUID = attr.ib()
    name: str = attr.ib()
    kind: str = attr.ib()
    owner: bool = attr.ib(default=False)
    workspaces: List[Workspace] = attr.ib(default=None)


@attr.s
class User:
    id: UUID = attr.ib()
    username: str = attr.ib()
    name: str = attr.ib()
    email: str = attr.ib()
    is_authenticated: bool = attr.ib()
    is_active: bool = attr.ib()
    is_anonymous: bool = attr.ib()
    orgs: List[Organization] = attr.ib(default=None)
    workspaces: List[Workspace] = attr.ib(default=None)

    def get_id(self) -> str:
        return str(self.id)

    def get_org(self, org_id: Union[UUID, str]) -> Organization:
        for o in self.orgs:
            if o.id == org_id:
                return o

    def get_workspace(self, workspace_id: Union[UUID, str]) -> Workspace:
        for w in self.workspaces:
            if w.id == workspace_id:
                return w



# -*- coding: utf-8 -*-
from uuid import UUID

from flask import abort, Blueprint, jsonify, request
from flask_login import current_user, login_required
from marshmallow import ValidationError

from chaosplt_account.schemas import new_org_schema, org_schema, \
    link_workspace_schema, org_schema_short, orgs_schema_tiny, \
    workspaces_schema

__all__ = ["api"]

api = Blueprint("org", __name__)


@api.route('', methods=['GET'])
@login_required
def list_all():
    orgs = request.services.account.org.list_all()
    return orgs_schema_tiny.jsonify(orgs)


@api.route('', methods=['POST'])
@login_required
def new():
    user_id = current_user.id
    try:
        payload = new_org_schema.load(request.json)
    except ValidationError as err:
        return jsonify(err.messages), 422

    org_name = payload["name"]

    has_org = request.services.account.org.has_org_by_name(org_name)
    if has_org:
        return jsonify({
            "message": "Name already used"
        }), 409

    org = request.services.account.org.create(org_name, user_id)
    return org_schema.jsonify(org), 201


@api.route('<uuid:org_id>', methods=['GET'])
@login_required
def get(org_id: UUID):
    org = request.services.account.org.get(org_id)
    if not org:
        return abort(404)
    return org_schema_short.jsonify(org)


@api.route('<uuid:org_id>', methods=['DELETE'])
@login_required
def delete(org_id: UUID):
    request.services.account.org.delete(org_id)
    return "", 204


@api.route('<uuid:user_id>/workspaces', methods=['GET'])
@login_required
def get_user_workspaces(org_id: UUID):
    org = request.services.account.org.get(org_id)
    if not org:
        return abort(404)
    return workspaces_schema.jsonify(org.workspaces)


@api.route('<uuid:org_id>/workspaces/<uuid:workspace_id>', methods=['PUT'])
@login_required
def link_workspace_to_org(org_id: UUID, workspace_id: UUID):
    try:
        payload = link_workspace_schema.load(request.json)
    except ValidationError as err:
        return jsonify(err.messages), 422

    org = request.services.account.org.get(org_id)
    if not org:
        return abort(404)

    workspace = request.services.account.workspace.get(workspace_id)
    if not workspace:
        return abort(404)

    request.services.account.org.add_workspace(
        org_id, workspace_id, owner=payload["owner"])
    return "", 204


@api.route('<uuid:org_id>/organizations/<uuid:workspace_id>',
           methods=['DELETE'])
@login_required
def unlink_workspace_from_org(org_id: UUID, workspace_id: UUID):
    org = request.services.account.org.get(org_id)
    if not org:
        return abort(404)

    workspace = request.services.account.workspace.get(workspace_id)
    if not workspace:
        return abort(404)

    request.services.account.org.remove_org(org_id, workspace_id)
    return "", 204



from typing import Any, Dict, NoReturn

from chaosplt_relational_storage import get_storage, \
    configure_storage, release_storage
import pkg_resources

from .concrete import OrgService, RegistrationService, UserService, \
    WorkspaceService
from .interface import BaseAccountStorage

__all__ = ["initialize_storage", "shutdown_services", "AccountStorage"]


class AccountStorage(BaseAccountStorage):
    def __init__(self, config: Dict[str, Any]):
        self.driver = get_storage(config)
        configure_storage(self.driver)

        user = UserService(self.driver)
        org = OrgService(self.driver)
        workspace = WorkspaceService(self.driver)
        registration = RegistrationService(self.driver)

        BaseAccountStorage.__init__(
            self, user, org, workspace, registration)

    def release(self) -> NoReturn:
        release_storage(self.driver)


def initialize_storage(config: Dict[str, Any]) -> BaseAccountStorage:
    for plugin in pkg_resources.iter_entry_points('chaosplatform.storage'):
        if plugin.name == "account":
            service_class = plugin.load()
            return service_class(config)

    return AccountStorage(config)


def shutdown_storage(storage: BaseAccountStorage) -> NoReturn:
    storage.release()
